/**
 * Classe resposaável por conectar no Asterisk, higienizar os dados e fazer um de/para dos eventos.
 */
const EventEmitter = require('events');
class MyEmitter extends EventEmitter { }
const emitter = new MyEmitter();
var namiLib = require("nami");

const log = require("node-live-log")("AsteriskConn.js");

/**
 * Lista de Eventos de=>para. Na esquerda o nome do avento que o asterisk emite, da direita o nome do evento que essa classe emite.
 * 
 * O Array abaixo e a função limpa estão fora da classe pois são estaticos e globais e seria um saco ficar copiando eles antes this.nami.on para poder usar dentro das funções de callback;;
 */
const Eventos = {
    QueueMemberAdded: 'AgenteAdicionado',
    QueueMemberRemoved: 'AgenteRemovido',
    QueueMemberPause: 'AgentePausa',
    QueueMemberStatus: 'AgenteStatus',
    QueueCallerJoin: 'LigacaoEntrouNaFila',
    QueueCallerAbandon: 'LigacaoAbandonouNaFila',
    QueueCallerLeave: 'LigacaoSaiuDaFila',
    AgentCalled: 'AgenteTocando',
    AgentConnect: 'AgenteAtendeu',
    AgentRingNoAnswer: 'AgenteNaoAtendeu',
    AgentComplete: 'AgenteDesligou',
    AgentDump: 'AgenteDump',
    Newstate: 'LigacaoStatus',
    BlindTransfer: 'TransferenciaDireta',
    AttendedTransfer: 'TransferenciaConsulta',
    Hangup: 'LigacaoFim',
    FullyBooted: 'ConexaoAmiInciada'
};

// Remove dados inuteis dos eventos
const limpa = (response) => {
    const { lines, EOL, variables, privilege, event, ...ev } = response;
    return ev;
}

module.exports = class AsteriskConn {
    constructor() {
        this.nami = new (require("nami").Nami)(
            {
                host: process.env.AMI_HOST || 'localhost',
                port: process.env.AMI_PORT || 5038,
                username: process.env.AMI_USER || 'asterclic',
                secret: process.env.AMI_PASS || 'aster123',
                logLevel: 0
            }
        );
        this.nami.logLevel = 0;
        this.nami.on('namiConnected', function (event) {
            log.info('[namiConnected] Conectado ao Asterisk');
            emitter.emit('ConnectionOpen', event);
        });
        const vm = this;
        this.nami.on('namiConnectionClose', function (data) {
            log.warn('[namiConnectionClose] Perdi a conexão com o Asterisk, Reconnecting em 5s...');
            setTimeout(function () { vm.nami.open(); }, 5000);
        });

        this.nami.on('namiInvalidPeer', function (data) {
            log.warn('[namiInvalidPeer] Tem certeza que está tentando conectar em um AMI?');
            //process.exit();
        });
        this.nami.on('namiLoginIncorrect', function () {
            log.warn('[namiLoginIncorrect] Usuário ou senha recusado.');
            //process.exit();
        });

        this.nami.on('namiEvent', function (evCoco) {
            // Limpa e higeniza essa bosta
            const { lines, EOL, variables, privilege, event, ...ev } = evCoco;            
            try {
                if (!Eventos[event]) return;
                // Só gera log se for um evento mapeado
                log.debug(`[AMI_EVENTO] ${event}:`, ev);
                emitter.emit(Eventos[event], ev);
            } catch (ex) {
                log.warn(`AMI_EVENTO: Erro ao tratar evento ${event}, ERRO: `, ex);
            }
        });

        this.nami.open();
    }
    on(event, fn) {
        return emitter.addListener(event, fn);
    }

    /**
     * Obtem a lista das filas e agentes carregadas pelo Asterisk.
     * Esse comando do asterisk retorna um evento com um array de eventos que podem ser de 2 tipos: QueueParams = Dados da fila e QueueMember = Dados do agente logado na fila.
     * @param {String} fila Opcional, Nome da fila. Se não definido traz de todas as filas.
     * @param {Function} reply Função para processar o retorno do comando (Callback).
     */
    getFilas(fila, reply) {
        this.nami.send(new namiLib.Actions.QueueStatus(fila), function (response) {
            if (typeof reply === 'function') {
                reply(limpa(response));
            }
        });
    }

    /**
     * Adiciona um agente a uma ou várias filas.
     * request.ramal (String, Obrigatório): Ramal do agente
     * request.fila  (String ou Array, Obrigatório): Fila(s)
     * request.nome (String, Obrigatório): Nome do agente
     * request.pausado (Int, Opcional, Default 0): 0 - Adiciona como disponível, 1 - Adiciona em pausa.
     * request.penalty (Int, Opcional): Número de penalidades
     * @param {Object} request Objeto com dados da requisição
     * @returns Promise
     */
    AdicionarAgente(request) {
        //console.log('---------------- AdicionarAgente ------------------')
        //console.log(request)
        if (Array.isArray(request.fila)) {
            let parray = [];
            for (let i = 0; i < request.fila.length; i++) {
                let fila = request.fila[i];
                const p = new Promise((resolve, reject) => {
                    const qa = new namiLib.Actions.QueueAdd(request.ramal, fila, request.pausado, request.nome, request.penalty);
                    this.nami.send(qa, (resposta) => {
                        //console.log('------ Resposta ----------', resposta)
                        resposta.ramal = request.ramal;
                        resposta.fila = fila;
                        resolve(limpa(resposta));
                    });
                });
                parray.push(p);
            }
            return Promise.all(parray);
        }
        else {
            return new Promise((resolve, reject) => {
                const qa = new namiLib.Actions.QueueAdd(request.ramal, request.fila, request.pausado, request.nome, request.penalty);
                this.nami.send(qa, (resposta) => {
                    //console.log('------ Resposta ----------', resposta)
                    resposta.ramal = request.ramal;
                    resposta.fila = request.fila;
                    resolve(limpa(resposta));
                });
            });
        }
    };

    /**
     * Remove um agente de uma ou várias filas.
     * request.ramal (String, Obrigatório): Ramal do agente
     * request.fila  (String ou Array, Obrigatório): Fila(s)
     * @param {Object} request Objeto com dados da requisição
     * @returns Promise
     */
    RemoverAgente(request) {
        if (Array.isArray(request.fila)) {
            let parray = [];
            for (let i = 0; i < request.fila.length; i++) {
                let fila = request.fila[i];
                const p = new Promise((resolve, reject) => {
                    const qa = new namiLib.Actions.QueueRemove(request.ramal, fila);
                    this.nami.send(qa, (resposta) => {
                        resposta.ramal = request.ramal;
                        resposta.fila = fila;
                        resolve(limpa(resposta));
                    });
                });
                parray.push(p);
            }
            return Promise.all(parray);
        }
        else {
            return new Promise((resolve, reject) => {
                const qa = new namiLib.Actions.QueueRemove(request.ramal, request.fila);
                this.nami.send(qa, (resposta) => {
                    resposta.ramal = request.ramal;
                    resposta.fila = request.fila;
                    resolve(limpa(resposta));
                });
            });
        }
    };

    /**
     * Pausa um agente em todas as filas.
     * @param {Object} request 
     * @returns Promise
     */
    PausarAgente(request) {
        return new Promise((resolve, reject) => {
            const qa = new namiLib.Actions.QueuePause(request.ramal, undefined, request.motivo);
            this.nami.send(qa, (resposta) => {
                resposta.ramal = request.ramal;
                resolve(limpa(resposta));
            });
        });
    };

    /**
     * Tira um agente da pausa.
     * @param {Object} request 
     * @returns Promise
     */
    TirarPausaAgente(request, reply) {
        return new Promise((resolve, reject) => {
            const qa = new namiLib.Actions.QueueUnpause(request.ramal);
            this.nami.send(qa, (resposta) => {
                resposta.ramal = request.ramal;
                resolve(limpa(resposta));
            });
        });
    };

    /**
     * Disca para um destino. Funciona como o discador preview, primeiro entrega para o ramal e depois disca para o destino.
     * Só retorna falha se não conseguir criar o comando originate, qualquer outra falha precisa ser monitorada pelo evento OriginateResponse
     * @param {Object} request 
     * @returns Promise
     */
    Discar(request) {
        const ac = new namiLib.Actions.Originate();
        ac.channel = request.origem;
        ac.exten = request.destino;
        ac.callerid = request.destino_nome ? request.destino_nome + " <" + request.destino + ">" : request.destino;
        ac.priority = 1;
        ac.context = request.rota;
        ac.timeout = request.max_chamar ? request.max_chamar * 1000 : 30 * 1000;
        ac.channelid = new Date().getTime();
        ac.async = true;
        let ar = [];
        ar.push('__SIPADDHEADER51=X-UNIQUE-ID:' + ac.channelid);
        if (request.campos && Object.keys(request.campos).length > 0) {
            let id = 1;
            Object.keys(request.campos).forEach(function (key) {
                ar.push('__SIPADDHEADER5' + id + '=X-' + key + ':' + encodeURIComponent(request.campos[key]));
                ++id;
            });
        }
        ac.variable = ar.join(",");

        return new Promise((resolve, reject) => {
            this.nami.send(ac, (resposta) => {
                resposta.callid = ac.channelid;
                resolve(limpa(resposta));
            });
        });
    };

    /**
     * Adiciona um destino a uma sala de conferência. Disca para o destino e quando ele atender coloca ele na sala.
     * @param {Object} request 
     * @returns 
     */
    Conferencia(request) {
        const ac = new namiLib.Actions.Originate();
        ac.channel = "Local/" + request.destino + "@" + request.rota;
        ac.application = "Confbridge";
        ac.data = request.sala + ",conf_webrtc,user_webrtc";
        ac.timeout = request.max_chamar ? request.max_chamar * 1000 : 30 * 1000;
        ac.channelid = new Date().getTime();
        ac.async = true;
        let ar = [];
        ar.push('__SIPADDHEADER51=X-UNIQUE-ID:' + ac.channelid);
        if (request.campos && Object.keys(request.campos).length > 0) {
            let id = 1;
            Object.keys(request.campos).forEach(function (key) {
                ar.push('__SIPADDHEADER5' + id + '=X-' + key + ':' + encodeURIComponent(request.campos[key]));
                ++id;
            });
        }
        ac.variable = ar.join(",");

        return new Promise((resolve, reject) => {
            this.nami.send(ac, (resposta) => {
                resposta.callid = ac.channelid;
                resolve(limpa(resposta));
            });
        });
    };

    /**
     * Desliga uma ligação por um número de destino ou pelo canal.
     * @param {Object} request 
     * @returns 
     */
    Desligar(request) {
        const ac = new namiLib.Actions.Hangup();
        ac.channel = request.destino ? '/^' + request.destino + '-.*$/' : request.canal;
        ac.cause = request.causa ? request.causa : 16;

        return new Promise((resolve, reject) => {
            this.nami.send(ac, (resposta) => {
                resposta.callid = ac.channelid;
                resolve(limpa(resposta));
            });
        });
    };
}